# -*- coding: utf-8 -*-

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models import Manager
from django.db.models.base import ModelBase
from django.utils.six import with_metaclass

from . import settings
from .exceptions import AlreadyRegistered, DescendantRegistered, NotRegistered
from .fields import TranslatedVirtualField, TranslationJSONField
from .manager import MultilingualManager, transform_translatable_fields


class FieldsAggregationMetaClass(type):
    '''
    Metaclass to handle custom inheritance of fields between classes.
    '''
    def __new__(cls, name, bases, attrs):
        attrs['fields'] = set(attrs.get('fields', ()))
        for base in bases:
            if isinstance(base, FieldsAggregationMetaClass):
                attrs['fields'].update(base.fields)
        attrs['fields'] = tuple(attrs['fields'])
        return super(FieldsAggregationMetaClass, cls).__new__(cls, name, bases, attrs)


class TranslationOptions(with_metaclass(FieldsAggregationMetaClass, object)):
    '''
    Translatable fields are declared by registering a model using
    ``TranslationOptions`` class with appropriate ``fields`` attribute.
    Model-specific fallback values and languages can also be given as class
    attributes.

    Options instances hold info about translatable fields for a model and its
    superclasses. The ``local_fields`` and ``fields`` attributes are mappings
    from fields to sets of their translation fields; ``local_fields`` contains
    only those fields that are handled in the model's database table (those
    inherited from abstract superclasses, unless there is a concrete superclass
    in between in the inheritance chain), while ``fields`` also includes fields
    inherited from concrete supermodels (giving all translated fields available
    on a model).

    ``related`` attribute inform whether this model is related part of some relation
    with translated model. This model may be not translated itself.
    ``related_fields`` contains names of reverse lookup fields.
    '''
    required_languages = ()

    def __init__(self, model):
        '''
        Create fields dicts without any translation fields.
        '''
        self.model = model
        self.registered = False
        self.related = False
        self.local_fields = dict((f, set()) for f in self.fields)
        self.fields = dict((f, set()) for f in self.fields)
        self.related_fields = []

    def validate(self, model):
        '''
        Perform validation of `TranslationOptions`.
        '''
        for field in self.fields.keys():
            try:
                self.model._meta.get_field(field)
            except FieldDoesNotExist:
                raise ImproperlyConfigured(
                    'Attribute {}.fields contains an item "{}", which is not a field (missing a comma?).'.format(
                        self.__class__.__name__,
                        field
                    )
                )

        # TODO: apply more validation to the options
        if self.required_languages:
            if isinstance(self.required_languages, (tuple, list)):
                self._check_languages(self.required_languages, model)
            else:
                self._check_languages(self.required_languages.keys(), model)

                for fieldnames in self.required_languages.values():
                    for field in fieldnames:
                        if field not in self.fields:
                            raise ImproperlyConfigured(
                                'Fieldname "{}" in required_languages which is not '
                                'defined as translatable for Model "{}".'.format(field, model.__name__)
                            )

    def _check_languages(self, languages, model):
        valid_languages = list(settings.AVAILABLE_LANGUAGES) + list((settings.DEFAULT_LANGUAGE, ))
        for l in languages:
            if l not in valid_languages:
                raise ImproperlyConfigured(
                    'Language "{}" is in required_languages on Model "{}" but '
                    'not in settings.AVAILABLE_LANGUAGES.'.format(l, model.__name__)
                )

    def update(self, other):
        '''
        Update with options from a superclass.
        '''
        if other.model._meta.abstract:
            self.local_fields.update(other.local_fields)
        self.fields.update(other.fields)

    def add_translation_field(self, field, translation_field):
        '''
        Add a new translation field to both fields dicts.
        '''
        self.local_fields[field].add(translation_field)
        self.fields[field].add(translation_field)

    def get_field_names(self):
        '''
        Return name of all fields that can be used in filtering.
        '''
        return list(self.fields.keys()) + self.related_fields

    def __str__(self):
        local = tuple(self.local_fields.keys())
        inherited = tuple(set(self.fields.keys()) - set(local))
        return '%s: %s + %s' % (self.__class__.__name__, local, inherited)


def raise_if_field_exists(model, field_name):
    if not hasattr(model, field_name):
        return

    # Check if are not dealing with abstract field inherited.
    for cls in model.__mro__:
        if hasattr(cls, '_meta') and cls.__dict__.get(field_name, None):
            cls_opts = translator._get_options_for_model(cls)
            if not cls._meta.abstract or field_name not in cls_opts.local_fields:
                raise ValueError(
                    'Error adding translation field. Model "{}" already contains '
                    'a field named "{}".'.format(
                        model._meta.object_name, field_name
                    )
                )


def add_translation_field(model):
    '''
    Monkey patches the original model class to provide the `i18n` field.
    '''
    # field to store the translations in
    raise_if_field_exists(model, 'i18n')
    model.add_to_class('i18n', TranslationJSONField())


def add_virtual_fields(model, opts):
    '''
    Adds newly created translation fields to the given translation options.
    '''
    # proxy fields to assign and get values from.
    for field_name in opts.local_fields.keys():
        # first, add a `<original_field>_i18n` virtual field to get the currently
        # active translation for a field
        field = TranslatedVirtualField(
            original_field=field_name,
            blank=True,
            null=True,
            editable=False  # disable in admin
        )

        raise_if_field_exists(model, field.get_field_name())
        field.contribute_to_class(model, field.get_field_name())

        # add a virtual field pointing to the original field with name
        # <orignal_field>_<DEFAULT_LANGUAGE>
        field = TranslatedVirtualField(
            original_field=field_name,
            blank=True,
            null=True,
            editable=False,
            language=settings.DEFAULT_LANGUAGE
        )
        raise_if_field_exists(model, field.get_field_name())
        field.contribute_to_class(model, field.get_field_name())

        # now, for each language, add a virtual field to get the tranlation for
        # that specific langauge
        # <original_field>_<language>
        for language in list(settings.AVAILABLE_LANGUAGES):
            blank_allowed = language not in opts.required_languages
            field = TranslatedVirtualField(
                original_field=field_name,
                language=language,
                blank=blank_allowed,
                null=blank_allowed
            )
            raise_if_field_exists(model, field.get_field_name())
            field.contribute_to_class(model, field.get_field_name())


def has_custom_queryset(manager):
    '''
    Check whether manager (or its parents) has declared some custom get_queryset method.
    '''
    return getattr(manager, 'get_queryset', None) != getattr(Manager, 'get_queryset', None)


def add_manager(model):
    '''
    Monkey patches the original model to use MultilingualManager instead of
    default managers (not only ``objects``, but also every manager defined and inherited).

    Custom managers are merged with MultilingualManager.
    '''
    if model._meta.abstract:
        return

    def patch_manager_class(manager):
        if isinstance(manager, MultilingualManager):
            return
        if manager.__class__ is Manager:
            manager.__class__ = MultilingualManager
        else:
            class NewMultilingualManager(MultilingualManager, manager.__class__):
                use_for_related_fields = getattr(
                    manager.__class__, "use_for_related_fields", not has_custom_queryset(manager))
                _old_module = manager.__module__
                _old_class = manager.__class__.__name__

                def deconstruct(self):
                    return (
                        False,  # as_manager
                        '%s.%s' % (self._old_module, self._old_class),  # manager_class
                        None,  # qs_class
                        self._constructor_args[0],  # args
                        self._constructor_args[1],  # kwargs
                    )

            manager.__class__ = NewMultilingualManager

    managers = model._meta.local_managers
    for current_manager in managers:
        prev_class = current_manager.__class__
        patch_manager_class(current_manager)
        if model._default_manager.__class__ is prev_class:
            # Normally model._default_manager is a reference to one of model's managers
            # (and would be patched by the way).
            # However, in some rare situations (mostly proxy models)
            # model._default_manager is not the same instance as one of managers, but it
            # share the same class.
            model._default_manager.__class__ = current_manager.__class__
    patch_manager_class(model._base_manager)
    if hasattr(model._meta, "_expire_cache"):
        model._meta._expire_cache()


def patch_constructor(model):
    '''
    Monkey patches the original model to rewrite fields names in __init__
    '''
    old_init = model.__init__

    def patched_init(self, *args, **kwargs):
        old_init(self, *args, **transform_translatable_fields(self.__class__, kwargs))
    model.__init__ = patched_init


class Translator(object):
    '''
    A Translator object encapsulates an instance of a translator. Models are
    registered with the Translator using the register() method.
    '''
    create_virtual_fields = True

    def __init__(self):
        # All seen models (model class -> ``TranslationOptions`` instance).
        self._registry = {}

    def disable_create_virtual_fields(self):
        self.create_virtual_fields = False

    def register(self, model_or_iterable, opts_class=None, **options):
        '''
        Registers the given model(s) with the given translation options.

        The model(s) should be Model classes, not instances.

        Fields declared for translation on a base class are inherited by
        subclasses. If the model or one of its subclasses is already
        registered for translation, this will raise an exception.
        '''
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]

        for model in model_or_iterable:
            # Ensure that a base is not registered after a subclass (_registry
            # is closed with respect to taking bases, so we can just check if
            # we've seen the model).
            if model in self._registry:
                if self._registry[model].registered:
                    raise AlreadyRegistered(
                        'Model "{}" is already registered for translation'.format(
                            model.__name__
                        )
                    )
                else:
                    descendants = [d.__name__ for d in self._registry.keys()
                                   if issubclass(d, model) and d != model]
                    print(descendants)
                    raise DescendantRegistered(
                        'Model "%s" cannot be registered after its subclass'
                        ' "%s"' % (model.__name__, descendants[0]))

            # Find inherited fields and create options instance for the model.
            opts = self._get_options_for_model(model, opts_class, **options)

            # If an exception is raised during registration, mark model as not-registered
            try:
                self._register_single_model(model, opts)
            except Exception:
                self._registry[model].registered = False
                raise

    def _register_single_model(self, model, opts):
        # Now, when all fields are initialized and inherited, validate configuration.
        opts.validate(model=model)

        # Mark the object explicitly as registered -- registry caches
        # options of all models, registered or not.
        opts.registered = True

        # Add translation fields to the model.
        if not model._meta.proxy:
            add_translation_field(model)

            if self.create_virtual_fields:
                add_virtual_fields(model, opts)

        if self.create_virtual_fields:
            # Set MultilingualManager
            add_manager(model)

            # Patch __init__ to rewrite fields
            patch_constructor(model)

    def unregister(self, model_or_iterable):
        '''
        Unregisters the given model(s).

        If a model isn't registered, this will raise NotRegistered. If one of
        its subclasses is registered, `DescendantRegistered` will be raised.
        '''
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            # Check if the model is actually registered (``get_options_for_model``
            # throws an exception if it's not).
            self.get_options_for_model(model)
            # Invalidate all submodels options and forget about
            # the model itself.
            for desc, desc_opts in list(self._registry.items()):
                if not issubclass(desc, model):
                    continue
                if model != desc and desc_opts.registered:
                    # Allowing to unregister a base would necessitate
                    # repatching all submodels.
                    raise DescendantRegistered(
                        'You need to unregister descendant "{}" before'
                        ' unregistering its base "{}"'.format(
                            desc.__name__, model.__name__
                        )
                    )
                del self._registry[desc]

    def get_registered_models(self, abstract=True):
        '''
        Returns a list of all registered models, or just concrete
        registered models.
        '''
        return [model for (model, opts) in self._registry.items()
                if opts.registered and (not model._meta.abstract or abstract)]

    def _get_options_for_model(self, model, opts_class=None, **options):
        '''
        Returns an instance of translation options with translated fields
        defined for the ``model`` and inherited from superclasses.
        '''
        if model not in self._registry:
            # Create a new type for backwards compatibility.
            opts = type('%sTranslationOptions' % model.__name__,
                        (opts_class or TranslationOptions,), options)(model)

            # Fields for translation may be inherited from abstract
            # superclasses, so we need to look at all parents.
            for base in model.__bases__:
                if not hasattr(base, '_meta'):
                    # Things without _meta aren't functional models, so they're
                    # uninteresting parents.
                    continue
                opts.update(self._get_options_for_model(base))

            # Cache options for all models -- we may want to compute options
            # of registered subclasses of unregistered models.
            self._registry[model] = opts

        return self._registry[model]

    def get_options_for_model(self, model):
        '''
        Thin wrapper around ``_get_options_for_model`` to preserve the
        semantic of throwing exception for models not directly registered.
        '''
        opts = self._get_options_for_model(model)
        if not opts.registered and not opts.related:
            raise NotRegistered(
                'The model "{}" is not registered for translation'.format(model.__name__)
            )
        return opts


# This global object represents the singleton translator object
translator = Translator()



from django.utils import six
from django.utils.encoding import force_text
from django.utils.functional import lazy
from django.utils.translation import get_language as _get_language

from modeltrans import settings


def get_language():
    '''
    Return an active language code that is guaranteed to be in
    settings.LANGUAGES (Django does not seem to guarantee this for us).
    '''
    lang = _get_language()
    if lang is None:  # Django >= 1.8
        return settings.DEFAULT_LANGUAGE
    if lang not in settings.AVAILABLE_LANGUAGES and '-' in lang:
        lang = lang.split('-')[0]
    if lang in settings.AVAILABLE_LANGUAGES:
        return lang
    return settings.DEFAULT_LANGUAGE


def get_translation_fields(field):
    '''
    Returns a list of localized fieldnames for a given field.
    '''
    return [build_localized_fieldname(field, l) for l in settings.AVAILABLE_LANGUAGES]


def split_translated_fieldname(field_name):
    _pos = field_name.rfind('_')
    return (field_name[0:_pos], field_name[_pos + 1:])


def build_localized_fieldname(field_name, lang):
    if lang == 'id':
        # The 2-letter Indonesian language code is problematic with the
        # current naming scheme as Django foreign keys also add "id" suffix.
        lang = 'ind'
    return str('{}_{}'.format(field_name, lang.replace('-', '_')))


def _build_localized_verbose_name(verbose_name, lang):
    if lang == 'id':
        lang = 'ind'
    return force_text('%s [%s]') % (force_text(verbose_name), lang)


build_localized_verbose_name = lazy(_build_localized_verbose_name, six.text_type)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import AppConfig


class RegistrationConfig(AppConfig):
    name = 'modeltrans'
    verbose_name = 'Django modeltrans using a registry.'

    def ready(self):
        from modeltrans.models import handle_translation_registrations
        handle_translation_registrations()



# -*- coding: utf-8 -*-

from django.core.exceptions import FieldError
from django.db import models
from django.db.models import TextField
from django.db.models.expressions import RawSQL
from django.db.models.functions import Cast, Coalesce

from . import settings
from .utils import (build_localized_fieldname, get_language,
                    split_translated_fieldname)


def get_translatable_fields_for_model(Model):
    from modeltrans.translator import NotRegistered, translator
    try:
        return translator.get_options_for_model(Model).get_field_names()
    except NotRegistered:
        return None


def transform_translatable_fields(model, fields):
    '''
    Transform the kwargs for a <Model>.objects.create() or <Model>()
    to allow passing translated field names.

    Arguments:
        fields (dict): kwargs to a model __init__ or Model.objects.create() method
            for which the field names need to be translated to values in the i18n field
    '''

    ret = {
        'i18n': fields.get('i18n', {})
    }

    for field, value in fields.items():
        original_field, lang = split_translated_fieldname(field)

        if lang == settings.DEFAULT_LANGUAGE:
            if original_field in fields:
                raise ValueError(
                    'Attempted override of "{}" with "{}". '
                    'Only one of the two is allowed.'.format(original_field, field)
                )
            ret[original_field] = value
        elif original_field in get_translatable_fields_for_model(model):
            ret['i18n'][field] = value
        else:
            ret[field] = value

    return ret


class MultilingualQuerySet(models.query.QuerySet):
    def get_translatable_fields(self):
        return get_translatable_fields_for_model(self.model)

    def add_i18n_annotation(self, original_field, field_name, annotation_field_name=None, fallback=True):
        '''
        Add an annotation to the query to extract the translated version of a field
        from the jsonb field to allow filtering and ordering.

        Arguments:
            original_field (str): name of the original, untranslated field.
            field_name (str): name of the translated field to add the
                annotation for. For example `title_nl` will result in adding
                something like `i18n->>title_nl AS title_nl_annotation` to the Query.
            annotation_field_name (str): name of the annotation, if None (by default),
                `<original_field>_<lang>_annotation` will be used.
            fallback (bool): If `True`, `COALESCE` will be used to get the value
                of the original field if the requested translation is not
                available.

        Returns:
            the name of the annotation.
        '''
        assert field_name.startswith(original_field)

        if original_field not in self.get_translatable_fields():
            raise FieldError('Field ({}) is not defined as translatable'.format(original_field))

        if fallback:
            # fallback to the original untranslated field
            field = Coalesce(RawSQL('i18n->>%s', (field_name, )), original_field, output_field=TextField())
        else:
            field = Cast(RawSQL('i18n->>%s', (field_name, )), TextField())

        if annotation_field_name is None:
            annotation_field_name = '{}_annotation'.format(field_name)
        self.query.add_annotation(field, annotation_field_name)

        return annotation_field_name

    def create(self, **kwargs):
        '''
        Patch the create method to allow adding the value for a translated field
        using `Model.objects.create(..., title_nl='...')`.
        '''
        return super(MultilingualQuerySet, self).create(
            **transform_translatable_fields(self.model, kwargs)
        )

    def order_by(self, *field_names):
        '''
        Annotate translated fields before sorting.

        Examples:
         - sort on `-title_nl` will add an annotation for `title_nl`
         - sort on `title_i18n` will add an annotation for the current language

        The field names pointing to translated fields in the `field_names`
        argument will be replaced by their annotated versions.
        '''
        new_field_names = []

        for field in field_names:
            if '_' not in field:
                new_field_names.append(field)
                continue

            # remove descending prefix, not relevant for the annotation
            if field[0] == '-':
                field = field[1:]
                descending = True
            else:
                descending = False

            original_field, language = split_translated_fieldname(field)

            # sort by current language if <original_field>_i18n is requested
            if language == 'i18n':
                language = get_language()
                field = build_localized_fieldname(original_field, language)

            if language == settings.DEFAULT_LANGUAGE:
                sort_field_name = original_field
            else:
                sort_field_name = self.add_i18n_annotation(original_field, field, fallback=True)

            # re-add the descending prefix to the annotated field name
            if descending:
                sort_field_name = '-' + sort_field_name

            new_field_names.append(sort_field_name)

        return super(MultilingualQuerySet, self).order_by(*new_field_names)

    def rewrite_expression(self, lookup, value):
        requested_field_name = lookup
        query_type = ''

        # strip the query type
        if '__' in lookup:
            lookup = lookup[0:lookup.rfind('__')]
            query_type = requested_field_name[len(lookup):]

        original_field, language = split_translated_fieldname(lookup)
        if original_field not in self.get_translatable_fields():
            return requested_field_name, value

        if language == 'i18n':
            # search for current language, including fallback to
            # settings.DEFAULT_LANGUAGE
            language = get_language()
            lookup = build_localized_fieldname(original_field, language)
            fallback = True
        else:
            fallback = False

        if language == settings.DEFAULT_LANGUAGE:
            filter_field_name = original_field
        else:
            filter_field_name = self.add_i18n_annotation(original_field, lookup, fallback=fallback)

        # re-add query type
        filter_field_name += query_type

        return filter_field_name, value

    def rewrite_Q(self, q):
        if isinstance(q, models.Q):
            return models.Q._new_instance(
                list(self.rewrite_Q(child) for child in q.children),
                connector=q.connector,
                negated=q.negated
            )
        if isinstance(q, (list, tuple)):
            return self.rewrite_expression(*q)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        '''
        Annotate lookups for `filter()` and `exclude()`.

        Examples:
            - `title_nl__contains='foo'` will add an annotation for `title_nl`
            - `title_nl='bar'` will add an annotation for `title_nl`
            - `title_i18n='foo'` will add an annotation for `title_<language>`
              where `<language>` is the current active language.
            - `Q(title_nl__contains='foo') will add an annotation for `title_nl`

        In all cases, the field part of the field lookup will be changed to use
        the annotated verion.
        '''
        # TODO: handle F expressions in the righthand (value) side of filters

        # handle Q expressions
        new_args = []
        for arg in args:
            new_args.append(self.rewrite_Q(arg))

        # handle the kwargs
        new_kwargs = {}
        for field, value in kwargs.items():
            new_kwargs.update(dict((self.rewrite_expression(field, value), )))

        return super(MultilingualQuerySet, self)._filter_or_exclude(negate, *new_args, **new_kwargs)

    def _values(self, *fields, **expressions):
        '''
        Annotate lookups for `values()` and `values_list()`

        It must be possible to use `Blogs.objects.all().values_list('title_i18n', 'title_nl', 'title_en')`
        '''
        _fields = fields + tuple(expressions)

        for field in _fields:
            if '_' not in field:
                continue

            original_field, language = split_translated_fieldname(field)

            annotation_field_name = field
            fallback = False
            # current language if <original_field>_i18n is requested
            if language == 'i18n':
                language = get_language()
                field = build_localized_fieldname(original_field, language)
                fallback = True

            if language == settings.DEFAULT_LANGUAGE:
                # TODO: see if we can just do this with add_i18n_annotation()
                self.query.add_annotation(Cast(original_field, TextField()), field)
            else:
                self.add_i18n_annotation(
                    original_field,
                    field,
                    annotation_field_name=annotation_field_name,
                    fallback=fallback
                )

        return super(MultilingualQuerySet, self)._values(*fields, **expressions)


def multilingual_queryset_factory(old_cls, instantiate=True):
    '''Return a MultilingualQuerySet, or mix MultilingualQuerySet in custom QuerySets.'''
    if old_cls == models.query.QuerySet:
        NewClass = MultilingualQuerySet
    else:
        class NewClass(old_cls, MultilingualQuerySet):
            pass
        NewClass.__name__ = 'Multilingual%s' % old_cls.__name__
    return NewClass() if instantiate else NewClass


class MultilingualManager(models.Manager):
    use_for_related_fields = True

    def _patch_queryset(self, qs):
        qs.__class__ = multilingual_queryset_factory(qs.__class__, instantiate=False)
        return qs

    def get_queryset(self):
        '''
        This method is repeated because some managers that don't use super() or alter queryset class
        may return queryset that is not subclass of MultilingualQuerySet.
        '''
        qs = super(MultilingualManager, self).get_queryset()
        if isinstance(qs, MultilingualQuerySet):
            # Is already patched
            return qs
        return self._patch_queryset(qs)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

from collections import defaultdict

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Creates the datamigration the specified app'

    def add_arguments(self, parser):
        parser.add_argument('apps', nargs='+', type=str)

    def handle(self, *args, **options):
        from modeltrans.migration import (get_translatable_models,
                                          I18nMigration, get_translated_fields)

        models = get_translatable_models()

        apps = defaultdict(list)
        for model in models:
            apps[model._meta.app_label].append(model)

        for app in options['apps']:
            print('Create migration for app: ', app)
            migration = I18nMigration(app)

            for Model in apps[app]:
                translatable_fields = tuple(get_translated_fields(Model))

                print('added model "{}" with fields [{}]('.format(
                    Model.__name__,
                    str(translatable_fields)
                ))
                migration.add_model(Model.__name__, translatable_fields)

            migration.write()



#!/usr/bin/env python
import os
import re
import sys

from setuptools import find_packages, setup

# get version without importing
with open('modeltrans/__init__.py', 'rb') as f:
    VERSION = str(re.search('__version__ = \'(.+?)\'', f.read().decode('utf-8')).group(1))

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    message = '\nreleased [{version}](https://pypi.python.org/pypi/django-modeltrans/{version})'
    print(message.format(version=VERSION))
    sys.exit()

if sys.argv[-1] == 'tag':
    os.system("git tag -a v{} -m 'tagging v{}'".format(VERSION, VERSION))
    os.system('git push --tags && git push origin master')
    sys.exit()


setup(
    name='django-modeltrans',
    version=VERSION,
    description='Model translations in a jsonb field',

    author='Jan Pieter Waagmeester',
    author_email='jieter@zostera.nl',
    license='Simplified BSD',
    url='https://github.com/zostera/django-modeltrans/',

    packages=find_packages(exclude=['tests.*', 'tests', 'example.*', 'example']),
    include_package_data=True,  # declarations in MANIFEST.in

    install_requires=['Django>=1.10'],

    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
    ],
)





def autodiscover(create_virtual_fields=True):
    '''
    Auto-discover INSTALLED_APPS translation.py modules and fail silently when
    not present. This forces an import on them to register.
    Also import explicit modules.
    '''
    import os
    import sys
    import copy
    from django.utils.module_loading import module_has_submodule
    from modeltrans.translator import translator

    if not create_virtual_fields:
        translator.disable_create_virtual_fields()

    from importlib import import_module
    from django.conf import settings
    from django.apps import apps
    mods = [(app_config.name, app_config.module) for app_config in apps.get_app_configs()]

    for (app, mod) in mods:
        # Attempt to import the app's translation module.
        module = '%s.translation' % app
        before_import_registry = copy.copy(translator._registry)
        try:
            import_module(module)
        except:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            translator._registry = before_import_registry

            # Decide whether to bubble up this error. If the app just
            # doesn't have an translation module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'translation'):
                raise

    # In debug mode, print a list of registered models and pid to stdout.
    # Note: Differing model order is fine, we don't rely on a particular
    # order, as far as base classes are registered before subclasses.
    if settings.DEBUG:
        try:
            if sys.argv[1] in ('runserver', 'runserver_plus'):
                models = translator.get_registered_models()
                names = ', '.join(m.__name__ for m in models)
                print('modeltrans: Registered %d models for translation'
                      ' (%s) [pid: %d].' % (len(models), names, os.getpid()))
        except IndexError:
            pass


def handle_translation_registrations(create_virtual_fields=True):
    '''
    Ensures that any configuration of the TranslationOption(s) are handled when
    importing modeltranslation.
    This makes it possible for scripts/management commands that affect models
    but know nothing of modeltranslation.
    '''
    from modeltrans.settings import ENABLE_REGISTRATIONS

    if not ENABLE_REGISTRATIONS:
        # If the user really wants to disable this, they can, possibly at their
        # own expense. This is generally only required in cases where other
        # apps generate import errors and requires extra work on the user's
        # part to make things work.
        return

    # Trigger autodiscover, causing any TranslationOption initialization
    # code to execute.
    autodiscover(create_virtual_fields)



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# Generated by django-modeltrans {version} on {timestamp}
from __future__ import unicode_literals

import inspect
import sys

from django.core.exceptions import ImproperlyConfigured
from django.utils.timezone import now

from modeltrans import __version__ as VERSION
from modeltrans.utils import split_translated_fieldname

from .settings import DEFAULT_LANGUAGE

try:
    from modeltranslation.translator import translator
    DJANGO_MODELTRANSLATION_AVAILABLE = True
except ImportError:
    DJANGO_MODELTRANSLATION_AVAILABLE = False


def _raise_if_not_django_modeltranslation():
    if not DJANGO_MODELTRANSLATION_AVAILABLE:
        raise ImproperlyConfigured(
            'django-modeltranslation must be still installed when creating'
            'the modeltranslation -> modeltrans migrations.'
        )


def get_translatable_models():
    _raise_if_not_django_modeltranslation()
    return translator.get_registered_models()


def get_translated_fields(Model):
    '''
    Enumerates the translated fields for a model according to django-modeltranslation.
    For example: title_nl, title_en, title_fr, body_nl, body_en, body_fr
    '''
    _raise_if_not_django_modeltranslation()

    options = translator.get_options_for_model(Model)
    for original_field, fields in options.fields.items():
        for translated in fields:
            yield translated.name


def copy_translations(model, fields):
    for m in model.objects.all():
        m.i18n = {}
        for field in fields:
            value = getattr(m, field)
            if value is None:
                continue

            original_field, lang = split_translated_fieldname(field)

            if lang == DEFAULT_LANGUAGE:
                setattr(m, original_field, value)
            else:
                m.i18n[field] = value

        m.save()


class I18nMigration(object):
    helper_functions = (
        split_translated_fieldname,
        copy_translations,
    )

    def __init__(self, app):
        self.models = []
        self.app = app

    def get_helper_functions(self):
        for fn in self.helper_functions:
            yield inspect.getsource(fn)

    def add_model(self, model_name, fields):
        self.models.append(
            (model_name, fields)
        )

    def write(self, out=None):
        if out is None:
            out = sys.stdout

        from modeltrans import settings

        out.write(MIGRATION_TEMPLATE.format(
            version=VERSION,
            DEFAULT_LANGUAGE=settings.DEFAULT_LANGUAGE,
            timestamp=now().strftime('%Y-%m-%d %H:%M'),
            helpers='\n\n'.join(self.get_helper_functions()),
            todo=',\n        '.join([str(item) for item in self.models]),
            app=self.app,
            last_migration='# TODO: actually fetch this from somewhere'
        ))


MIGRATION_TEMPLATE = '''
# -*- coding: utf-8 -*-
# Generated by django-modeltrans {version} on {timestamp}

from __future__ import print_function, unicode_literals

from django.db import migrations

DEFAULT_LANGUAGE = '{DEFAULT_LANGUAGE}'


{helpers}

def forwards(apps, schema_editor):
    app = '{app}'
    todo = (
        {todo},
    )

    for model, fields in todo:
        Model = apps.get_model(app, model)

        copy_translations(Model, fields)


class Migration(migrations.Migration):

    dependencies = [
        ('{app}', '{last_migration}'),
    ]

    operations = [
        # The copying of values is (sort of) reversable by a no-op:
        #  - values are copied into i18n (which is not used by anything but django-modeltrans)
        #  - the default language is copied to the orignal field, which was not used
        #    with django-modeltrans.
        migrations.RunPython(forwards, migrations.RunPython.noop)
    ]
'''
